/** \ingroup rpmcli
 * \file lib/rpmchecksig.c
 * Verify the signature of a package.
 */

#include "system.h"

#include <ctype.h>

#include <rpm/rpmlib.h>			/* RPMSIGTAG & related */
#include <rpm/rpmpgp.h>
#include <rpm/rpmcli.h>
#include <rpm/rpmfileutil.h>	/* rpmMkTemp() */
#include <rpm/rpmsq.h>
#include <rpm/rpmts.h>
#include <rpm/rpmlog.h>
#include <rpm/rpmstring.h>
#include <rpm/rpmkeyring.h>

#include "rpmio/rpmio_internal.h" 	/* fdSetBundle() */
#include "lib/rpmlead.h"
#include "lib/header_internal.h"
#include "lib/rpmvs.h"
#include "lib/rpmts_internal.h"
#include "lib/rpmtypes.h"

#include "debug.h"

int _print_pkts = 0;

static int getRpmCfgKey()
{
	FILE *fp;
	int ret;
	char buf[128]={0};
	fp = fopen(RPMCFG, "rb");
	if(fp == NULL)
	{
		rpmlog_message(RPMLOG_WARNING, "rpmconfig.cfg file not exist!\n");
		return -1;
	}
	while(fgets(buf, 128, fp)){
		int len = strlen(buf);
		if(len > 0 && (buf[len-1]=='\n' || buf[len-1]=='\r'))
		{
			buf[--len]='\0';
		}
		if( !strcmp(buf, "rpmgmkey=1") )
		{
			ret = 0;	
			break;
		}				
		else{
			ret = -2;  //当存在配置文件，但value值不为1时，主要是为了让配置文件和value值都存在时，走国密，其他都按原来的走
		}
		memset(buf,0, sizeof(buf));
	}
	fclose(fp);
	return ret;
}

//.rpm.sig 文件： rpmsigkey=1 不验签；rpmsigkey=2 验签不通过时，停止安装；其余情况：验签不通过时，提示告警信息，放行继续安装
static int getRpmSigCfgKey()
{
	FILE *fp;
	int ret;
	char buf[128]={0};
	fp = fopen(RPMSIGFILE, "rb");
	if(fp == NULL)
	{
		rpmlog_message(RPMLOG_WARNING, "chk .rpm.sig file not exist!\n");
		return -1;
	}
	while(fgets(buf, 128, fp)){
		int len = strlen(buf);
		if(len > 0 && (buf[len-1]=='\n' || buf[len-1]=='\r'))
		{
			buf[--len]='\0';
		}
		if(!strcmp(buf, "rpmsigkey=1"))
		{
			ret = 1;	
			break;
		}			
		else if(!strcmp(buf, "rpmsigkey=2") ){
			ret = 2;
			break;
		}	
		else{
			ret = -2;  //当存在配置文件，但value值不为1或2时
		}
		memset(buf,0, sizeof(buf));
	}
	fclose(fp);
	return ret;
}

static int doImport(rpmts ts, const char *fn, char *buf, ssize_t blen)
{
    char const * const pgpmark = "-----BEGIN PGP ";
    size_t marklen = strlen(pgpmark);
    int res = 0;
    int keyno = 1;
    char *start = strstr(buf, pgpmark);

    do {
	uint8_t *pkt = NULL;
	uint8_t *pkti = NULL;
	size_t pktlen = 0;
	size_t certlen;
	
	/* Read pgp packet. */
	if (pgpParsePkts(start, &pkt, &pktlen) == PGPARMOR_PUBKEY) {
	    pkti = pkt;

	    /* Iterate over certificates in pkt */
	    while (pktlen > 0) {
		if (pgpPubKeyCertLen(pkti, pktlen, &certlen)) {
		    rpmlog(RPMLOG_ERR, _("%s: key %d import failed.\n"), fn,
			    keyno);
		    res++;
		    break;
		}

		/* Import pubkey certificate. */
		if (rpmtsImportPubkey(ts, pkti, certlen) != RPMRC_OK) {
		    rpmlog(RPMLOG_ERR, _("%s: key %d import failed.\n"), fn,
			    keyno);
		    res++;
		}
		pkti += certlen;
		pktlen -= certlen;
	    }
	} else {
	    rpmlog(RPMLOG_ERR, _("%s: key %d not an armored public key.\n"),
		   fn, keyno);
	    res++;
	}
	
	/* See if there are more keys in the buffer */
	if (start && start + marklen < buf + blen) {
	    start = strstr(start + marklen, pgpmark);
	} else {
	    start = NULL;
	}

	keyno++;
	free(pkt);
    } while (start != NULL);

    return res;
}

int rpmcliImportPubkeys(rpmts ts, ARGV_const_t argv)
{
    int res = 0;
    for (ARGV_const_t arg = argv; arg && *arg; arg++) {
	const char *fn = *arg;
	uint8_t *buf = NULL;
	ssize_t blen = 0;
	char *t = NULL;
	int iorc;

	/* If arg looks like a keyid, then attempt keyserver retrieve. */
	if (rstreqn(fn, "0x", 2)) {
	    const char * s = fn + 2;
	    int i;
	    for (i = 0; *s && isxdigit(*s); s++, i++)
		{};
	    if (i == 8 || i == 16) {
		t = rpmExpand("%{_hkp_keyserver_query}", fn+2, NULL);
		if (t && *t != '%')
		    fn = t;
	    }
	}

	/* Read the file and try to import all contained keys */
	iorc = rpmioSlurp(fn, &buf, &blen);
	if (iorc || buf == NULL || blen < 64) {
	    rpmlog(RPMLOG_ERR, _("%s: import read failed(%d).\n"), fn, iorc);
	    res++;
	} else {
	    res += doImport(ts, fn, (char *)buf, blen);
	}
	
	free(t);
	free(buf);
    }
    return res;
}

static int readFile(FD_t fd, char **msg)
{
    unsigned char buf[4*BUFSIZ];
    ssize_t count;

    /* Read the payload from the package. */
    while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0) {}
    if (count < 0)
	rasprintf(msg, _("Fread failed: %s"), Fstrerror(fd));

    return (count != 0);
}

struct vfydata_s {
    int seen;
    int bad;
    int verbose;
};

static int vfyCb(struct rpmsinfo_s *sinfo, void *cbdata)
{
    struct vfydata_s *vd = cbdata;
    vd->seen |= sinfo->type;
    if (sinfo->rc != RPMRC_OK)
	vd->bad |= sinfo->type;
    if (vd->verbose) {
	char *vsmsg = rpmsinfoMsg(sinfo);
	#if 0
	int rret = getRpmCfgKey();
	int newret = getRpmSigCfgKey();
	if(rret != 0){
		if(newret == 1){
			rpmlog(RPMLOG_NOTICE, "    %s\n", vsmsg);
			rpmlog_message(RPMLOG_WARNING, "warning sign,&&&&&&    %s\n", vsmsg);
		}
		else if(newret == 2){
			rpmlog(RPMLOG_NOTICE, "    %s\n", vsmsg);
			rpmlog_message(RPMLOG_WARNING, "force sign,&&&&&&    %s\n", vsmsg);
		}
		else{
			rpmlog_message(RPMLOG_WARNING, "gpg&&&&&&    %s\n", vsmsg);
		}
	}
	else{
		rpmlog_message(RPMLOG_WARNING, "guomi&&&&&&    %s\n", vsmsg);
	}
	#else
	rpmlog(RPMLOG_NOTICE, "    %s\n", vsmsg);
	#endif
	free(vsmsg);
    }
    return 1;
}

struct rpmvs_s {
    struct rpmsinfo_s *sigs;
    int nsigs;
    int nalloced;
    rpmVSFlags vsflags;
    rpmDigestBundle bundle;
    rpmKeyring keyring;
    int vfylevel;
};

typedef struct FDSTACK_s * FDSTACK_t;

struct FDSTACK_s {
    FDIO_t		io;
    void *		fp;
    int			fdno;
    int			syserrno;	/* last system errno encountered */
    const char 		*errcookie;	/* pointer to custom error string */

    FDSTACK_t		prev;
};

/** \ingroup rpmio
 * Cumulative statistics for a descriptor.
 */
typedef	struct {
    struct rpmop_s	ops[FDSTAT_MAX];	/*!< Cumulative statistics. */
} * FDSTAT_t;

/** \ingroup rpmio
 * The FD_t File Handle data structure.
 */

struct _FD_s {
    int		nrefs;
    int		flags;
#define	RPMIO_DEBUG_IO		0x40000000
    int		magic;
#define	FDMAGIC			0x04463138
    FDSTACK_t	fps;
    int		urlType;	/* ufdio: */

    char	*descr;		/* file name (or other description) */
    FDSTAT_t	stats;		/* I/O statistics */

    rpmDigestBundle digests;
};

rpmRC rpmpkgRead(struct rpmvs_s *vs, FD_t fd,
		hdrblob *sigblobp, hdrblob *blobp, char **emsg)
{

    char * msg = NULL;
    rpmRC xx, rc = RPMRC_FAIL; /* assume failure */
    hdrblob sigblob = hdrblobCreate();
    hdrblob blob = hdrblobCreate();
    rpmDigestBundle bundle = fdGetBundle(fd, 1); /* freed with fd */

    if ((xx = rpmLeadRead(fd, &msg)) != RPMRC_OK) {
	/* Avoid message spew on manifests */
	if (xx == RPMRC_NOTFOUND)
	    msg = _free(msg);
	rc = xx;
	goto exit;
    }

    /* Read the signature header. Might not be in a contiguous region. */
    if (hdrblobRead(fd, 1, 0, RPMTAG_HEADERSIGNATURES, sigblob, &msg))
	goto exit;

    rpmvsInit(vs, sigblob, bundle);

    /* Initialize digests ranging over the header */
    rpmvsInitRange(vs, RPMSIG_HEADER);

    /* Read the header from the package. */
    if (hdrblobRead(fd, 1, 1, RPMTAG_HEADERIMMUTABLE, blob, &msg))
	goto exit;

    /* Finalize header range */
    rpmvsFiniRange(vs, RPMSIG_HEADER);

    /* Fish interesting tags from the main header. This is a bit hacky... */
    rpmvsAppendTag(vs, blob, RPMTAG_PAYLOADDIGEST);

    /* If needed and not explicitly disabled, read the payload as well. */
    if (rpmvsRange(vs) & RPMSIG_PAYLOAD) {
	/* Initialize digests ranging over the payload only */
	rpmvsInitRange(vs, RPMSIG_PAYLOAD);

	if (readFile(fd, &msg))
	    goto exit;

	/* Finalize payload range */
	rpmvsFiniRange(vs, RPMSIG_PAYLOAD);
	rpmvsFiniRange(vs, RPMSIG_HEADER|RPMSIG_PAYLOAD);
    }

    if (sigblobp && blobp) {
	*sigblobp = sigblob;
	*blobp = blob;
	sigblob = NULL;
	blob = NULL;
    }
    rc = RPMRC_OK;

exit:
    if (emsg)
	*emsg = msg;
    else
	free(msg);
    hdrblobFree(sigblob);
    hdrblobFree(blob);
    return rc;
}

static int rpmpkgVerifySigs(rpmKeyring keyring, int vfylevel, rpmVSFlags flags,
			   FD_t fd, const char *fn)
{
    char *msg = NULL;
    struct vfydata_s vd = { .seen = 0,
			    .bad = 0,
			    .verbose = rpmIsVerbose(),
    };
    int rc;
    struct rpmvs_s *vs = rpmvsCreate(vfylevel, flags, keyring);

    rpmlog(RPMLOG_NOTICE, "%s:%s", fn, vd.verbose ? "\n" : "");

    rc = rpmpkgRead(vs, fd, NULL, NULL, &msg);

    if (rc)
	goto exit;
	
    rc = rpmvsVerify(vs, RPMSIG_VERIFIABLE_TYPE, vfyCb, &vd);

    if (!vd.verbose) {
	if (vd.seen & RPMSIG_DIGEST_TYPE) {
	    rpmlog(RPMLOG_NOTICE, " %s", (vd.bad & RPMSIG_DIGEST_TYPE) ?
					_("DIGESTS") : _("digests"));
	}
	if (vd.seen & RPMSIG_SIGNATURE_TYPE) {
	    rpmlog(RPMLOG_NOTICE, " %s", (vd.bad & RPMSIG_SIGNATURE_TYPE) ?
					_("SIGNATURES") : _("signatures"));
	}
	rpmlog(RPMLOG_NOTICE, " %s\n", rc ? _("NOT OK") : _("OK"));
    }
	

exit:
    if (rc && msg)
	rpmlog(RPMLOG_ERR, "%s: %s\n", Fdescr(fd), msg);
    rpmvsFree(vs);
    free(msg);
    return rc;
}

/* Wrapper around rpmkVerifySigs to preserve API */
int rpmVerifySignatures(QVA_t qva, rpmts ts, FD_t fd, const char * fn)
{
    int rc = 1; /* assume failure */
    if (ts && qva && fd && fn) {
	rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
	rpmVSFlags vsflags = rpmtsVfyFlags(ts);
	int vfylevel = rpmtsVfyLevel(ts);
	rc = rpmpkgVerifySigs(keyring, vfylevel, vsflags, fd, fn);
    	rpmKeyringFree(keyring);
    }
    return rc;
}

//-----------------weiliuhu 2024.01.29
static int closeFile(FD_t *fdp)
{
    if (fdp == NULL || *fdp == NULL)
	return 1;

    /* close and reset *fdp to NULL */
    (void) Fclose(*fdp);
    *fdp = NULL;
    return 0;
}

static int msgCb(struct rpmsinfo_s *sinfo, void *cbdata)
{
    char **msg = cbdata;
    if (sinfo->rc && *msg == NULL)
	*msg = rpmsinfoMsg(sinfo);
    return (sinfo->rc != RPMRC_FAIL);
}


static int checkPkg(FD_t fd, char **msg)
{
    int rc;
    struct rpmvs_s *vs = rpmvsCreate(RPMSIG_DIGEST_TYPE, 0, NULL);
    off_t offset = Ftell(fd);

    Fseek(fd, 0, SEEK_SET);
    rc = rpmpkgRead(vs, fd, NULL, NULL, msg);
    if (!rc)
	rc = rpmvsVerify(vs, RPMSIG_DIGEST_TYPE, msgCb, msg);
    Fseek(fd, offset, SEEK_SET);

    rpmvsFree(vs);
    return rc;
}
#include <gmssl/sm2.h>
#include <gmssl/x509_alg.h>
#include <gmssl/x509.h>
#include <gmssl/rand.h>
#include <gmssl/error.h>



static unsigned char *myreadFile(char *path, unsigned int *length) {
	FILE *fp;
	unsigned char *data;
	fp = fopen(path, "rb");
	if (fp == NULL) {
		rpmlog_message(RPMLOG_ERR, "  path=%s\n", path);
		return NULL;
	}
	//只获取rpm的文件部分数据
	fseek(fp, 96+8, SEEK_SET);
	uint8_t list[8];
	fread(list, 8, 1, fp);	
	int i;
	
	int x1;
	int x2;
	uint8_t k1[4] ={list[3],list[2],list[1], list[0]};
	memcpy(&x1, k1, 4);
	uint8_t k2[4] = {list[7],list[6],list[5],list[4]};
	memcpy(&x2, k2, 4);
	memset(list, 0, 8);
	
	//96+16+9*16+(4244+4)
	int newlen = 96+16+x1*16+x2+4 +8;
	fseek(fp, newlen, SEEK_SET);
	fread(list, 8, 1, fp);

	int entryNum;
	uint8_t m[4] ={list[3],list[2],list[1], list[0]};
	memcpy(&entryNum, m, 4);
	
	unsigned int dataNum;
	uint8_t t[4] = {list[7],list[6],list[5],list[4]};
	
	memcpy(&dataNum, t, 4);
	int offsetLen = 96+16+x1*16+(x2+4)+16 + entryNum*16 + dataNum;
	
	fseek(fp, 0, SEEK_END);
	unsigned int alllen = ftell(fp);
	unsigned int readlen = alllen - offsetLen;
	
	data = (unsigned char *)malloc((readlen + 1) * sizeof(char));
	fseek(fp, offsetLen, SEEK_SET);
	
	*length = fread(data, 1, readlen, fp);
	data[*length] = '\0';
	fclose(fp);
	return data;
}

static unsigned char *myreadFileEx(char *path, unsigned int *length) {
	unsigned char *data;
	FD_t fd = Fopen(path, "r.ufdio");
	if (fd == NULL || Ferror(fd)) {
	    rpmlog_message(RPMLOG_ERR, "%s: open failed: %s\n", path, Fstrerror(fd));
	    return -1;
	}
	Fseek(fd, 96+8, SEEK_SET);
	uint8_t list[8];
	ssize_t tmpSize = Fread(list, 8, 1, fd);
	int i;
	
	int x1;
	int x2;
	uint8_t k1[4] ={list[3],list[2],list[1], list[0]};
	memcpy(&x1, k1, 4);
	uint8_t k2[4] = {list[7],list[6],list[5],list[4]};
	memcpy(&x2, k2, 4);
	memset(list, 0, 8);
	
	//96+16+9*16+(4244+4)
	int newlen = 96+16+x1*16+x2+4 +8;
	
	Fseek(fd, newlen, SEEK_SET);
	Fread(list, 8, 1, fd);

	int entryNum;
	uint8_t m[4] ={list[3],list[2],list[1], list[0]};
	memcpy(&entryNum, m, 4);
	unsigned int dataNum;
	uint8_t t[4] = {list[7],list[6],list[5],list[4]};
	
	memcpy(&dataNum, t, 4);
	int offsetLen = 96+16+x1*16+(x2+4)+16 + entryNum*16 + dataNum;
	
	Fseek(fd, 0, SEEK_END);
	unsigned int alllen = Ftell(fd);
	unsigned int readlen = alllen - offsetLen;
	
	data = (unsigned char *)malloc((readlen + 1) * sizeof(char));
	Fseek(fd, offsetLen, SEEK_SET);
	
	*length = Fread(data, 1, readlen, fd);
	data[*length] = '\0';

	Fseek(fd, 0, SEEK_SET);  //为了yum 流程开始gm验证后，文件头到起始位置
	Fclose(fd);
	return data;
}

//对rpm包验签 by weiliuhu 2024.01.26
static int kmsVerifyForRPM(char *pubKeyPath, char *rpmbagPath, unsigned char *sig, size_t siglen)
{
	if(pubKeyPath == NULL || rpmbagPath == NULL)
	{
		rpmlog_message(RPMLOG_ERR, "[kmsVerifyForRPM] path is null! rpmbagPath=%s\n", rpmbagPath);
		return -1;
	}
	int ret;
	unsigned char dgst[32];
	unsigned char *bufcert=NULL;
	unsigned int filelen = 0;
	bufcert = myreadFileEx(rpmbagPath, &filelen);
	sm3_digest(bufcert, filelen, dgst);
	free(bufcert);

	SM2_KEY sm2_key;
	FILE *pubkeyfp;
	memset(&sm2_key, 0, sizeof(sm2_key));
	pubkeyfp = fopen(pubKeyPath,"rb");
	if(pubkeyfp == NULL)
	{
		rpmlog_message(RPMLOG_ERR, "fopen %s error!\n", RPMPUBKEY);
		return -1;
	}
	ret = sm2_public_key_info_from_pem(&sm2_key,pubkeyfp);
	if(ret != 1)
	{
		rpmlog_message(RPMLOG_ERR, "sm2_private_key_from_pem  error! ret=%d\n", ret);
		fclose(pubkeyfp);
		return -1;
	}
	fclose(pubkeyfp);
	if ((ret = sm2_verify(&sm2_key, dgst, sig, siglen)) != 1) {
		rpmlog_message(RPMLOG_ERR, "    SM2 Signature Verify: Failed\n", ret);
		return -1;
	} else {
		rpmlog_message(RPMLOG_WARNING, "    SM2 Signature Verify: Success\n");
	}
	return 0;
}

static int kmsSM3VerifyForRPM(unsigned char *strold, size_t lenold, unsigned char *strnew)
{
	int rc = -1;
	if(strold == NULL || strnew == NULL)
	{
		rpmlog_message(RPMLOG_ERR, " strold || strnew is null !!!\n");
		goto exit;
	}
	unsigned char dgst[32];
	sm3_digest(strold, lenold, dgst);
	rc = memcmp(dgst, strnew, 32);
exit:
	if(rc == 0){
		rpmlog_message(RPMLOG_WARNING, "   SM3 : Success!!! \n");
	}
	else{
		rpmlog_message(RPMLOG_ERR, "   SM3 : Failed!!! \n");
	}
	return rc;
}

static int readRPMHeader(char *rpmpath)
{
	int  rc;
	GMHeader sigh = NULL;
	char *msg = NULL;
	char *msgg = NULL;
	
	FD_t fd = Fopen(rpmpath, "r.ufdio");
	if (fd == NULL || Ferror(fd)) {
	    rpmlog_message(RPMLOG_ERR, " %s:  open failed: %s\n",  rpmpath, Fstrerror(fd));
	    return -1;
	}

	if ((rc = rpmLeadRead(fd, &msgg)) != RPMRC_OK) {
		rpmlog_message(RPMLOG_ERR, "%s: %s\n", rpmpath, msgg);
		goto exit;
	}
	
	rc = rpmReadSignature(fd, &sigh, &msg);
	if (rc != RPMRC_OK) {
		rpmlog_message(RPMLOG_ERR, " %s:  rpmReadSignature failed: %s\n",  rpmpath, (msg && *msg ? msg : "\n"));
		goto exit;
	}
	int idx = -1;
	for(int i=0;i<sigh->indexUsed;i++)
	{
		if(sigh->index[i].info.tag == RPMTAG_SM3HEADER)
		{
			idx = i;
			break;
		}
    	}
    	if(idx == -1)
    	{
    		rpmlog_message(RPMLOG_ERR, "no found sm3 entry from header\n");
    		rc = -1;
    		goto exit;
    	}
   	int idxsm2=-1;
	for(int i = 0; i< sigh->indexUsed; i++)
	{
		if(sigh->index[i].info.tag == RPMTAG_SM2HEADER)
		{
			idxsm2= i;
			//对sm2签名数据进行sm3，然后跟已存的sm3值做对比
			rc = kmsSM3VerifyForRPM(sigh->index[i].data, sigh->index[i].info.count, sigh->index[idx].data);
			if(rc != 0)
			{
				rpmlog_message(RPMLOG_ERR, " kmsSM3VerifyForRPM error, rpmpath=%s, tag=%d, count=%d \n", rpmpath,sigh->index[i].info.tag, sigh->index[i].info.count);
				goto exit;
			}
			
			rc = kmsVerifyForRPM(RPMPUBKEY, rpmpath, sigh->index[i].data, sigh->index[i].info.count);
			if(rc != 0)
			{
				rpmlog_message(RPMLOG_ERR, " kmsVerifyForRPM error, rpmpath=%s, tag=%d, count=%d \n",  rpmpath,sigh->index[i].info.tag, sigh->index[i].info.count);
				goto exit;
			}
			break;
		}
	}
	if(idxsm2 == -1)
	{
    		rpmlog_message(RPMLOG_ERR, "no found sm2 entry from header\n");
    		rc = -1;
    		goto exit;
    	}
exit:
    if (fd)	(void) closeFile(&fd);
    if(sigh){
    	headerFree(sigh);
     }
    if(msg){
    	free(msg);
     }
    if(msgg){
    	free(msgg);
     }
    return rc;
}

int rpmcliVerifySignatures(rpmts ts, ARGV_const_t argv)
{
    int rret = -1;
	const char *arg1;
	rret = getRpmCfgKey();
	if(rret == 0)
	{
		int newret = getRpmSigCfgKey();
		ARGV_const_t tmp = argv;
		while((arg1 = *tmp++) != NULL){
			if(newret == 1){//不验签
				rpmlog_message(RPMLOG_WARNING, "GM not conducting signature verify, newret=%d\n",  newret);
			}
			else if(newret == 2){
				rret = readRPMHeader(arg1);
				if(rret != 0){
					rpmlog(RPMLOG_ERR, "Package gm signature failed to verify!\n");
					return -1;
				}
			}
			else{//验签失败时，告警，继续安装
				rret = readRPMHeader(arg1);
				if(rret != 0){
					rpmlog(RPMLOG_WARNING, "Package gm signature failed to verify.\n");
				}
			}
		}
		return 0;
	}
	
    const char * arg;
    int res = 0;
    rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
    rpmVSFlags vsflags = rpmtsVfyFlags(ts);
    int vfylevel = rpmtsVfyLevel(ts);

    if (rpmcliQueryFlags & QUERY_DIGEST)
	vsflags |= _RPMVSF_NODIGESTS;
    if (rpmcliQueryFlags & QUERY_SIGNATURE)
	vsflags |= _RPMVSF_NOSIGNATURES;
    vsflags |= rpmcliVSFlags;
    if (rpmcliVfyLevelMask) {
	vfylevel &= ~rpmcliVfyLevelMask;
	rpmtsSetVfyLevel(ts, vfylevel);
    }
    
	if(getRpmCfgKey() == 0)  //去掉gpg的验证
    {
    	vfylevel = rpmtsVfyLevel(ts);
    	vfylevel &= ~RPMSIG_SIGNATURE_TYPE;
    	rpmtsSetVfyLevel(ts, vfylevel);
    }

    while ((arg = *argv++) != NULL) {
	FD_t fd = Fopen(arg, "r.ufdio");
	if (fd == NULL || Ferror(fd)) {
	    rpmlog(RPMLOG_ERR, _("%s: open failed: %s\n"), 
		     arg, Fstrerror(fd));
	    res++;
	} else if (rpmpkgVerifySigs(keyring, vfylevel, vsflags, fd, arg)) {
	    res++;
	}

	Fclose(fd);
	rpmsqPoll();
    }
    rpmKeyringFree(keyring);
    return res;
}
