/** \ingroup rpmcli
 * \file lib/rpmchecksig.c
 * Verify the signature of a package.
 */

#include "system.h"

#include <errno.h>
#include <sys/wait.h>
#include <popt.h>

#include <rpm/rpmlib.h>			/* RPMSIGTAG & related */
#include <rpm/rpmmacro.h>
#include <rpm/rpmpgp.h>
#include <rpm/rpmsign.h>
#include <rpm/rpmfileutil.h>	/* rpmMkTemp() */
#include <rpm/rpmlog.h>
#include <rpm/rpmstring.h>
#include <rpmio/rpmio_internal.h>

#include "lib/rpmlead.h"
#include "lib/signature.h"
#include "lib/rpmvs.h"
#include "sign/rpmsignfiles.h"
#include "lib/rpmtypes.h"

#include "debug.h"

typedef struct sigTarget_s {
    FD_t fd;
    const char *fileName;
    off_t start;
    rpm_loff_t size;
} *sigTarget;

static int closeFile(FD_t *fdp)
{
    if (fdp == NULL || *fdp == NULL)
	return 1;

    /* close and reset *fdp to NULL */
    (void) Fclose(*fdp);
    *fdp = NULL;
    return 0;
}

/**
 */
static int manageFile(FD_t *fdp, const char *fn, int flags)
{
    FD_t fd;
    const char *fmode;

    if (fdp == NULL || fn == NULL)	/* programmer error */
	return 1;

    /* open a file and set *fdp */
    if (*fdp == NULL && fn != NULL) {
	switch (flags & O_ACCMODE) {
	    case O_WRONLY:
		fmode = "w.ufdio";
		break;
	    case O_RDONLY:
		fmode = "r.ufdio";
		break;
	    default:
	    case O_RDWR:
		fmode = "r+.ufdio";
		break;
	}
	fd = Fopen(fn, fmode);
	if (fd == NULL || Ferror(fd)) {
	    rpmlog(RPMLOG_ERR, _("%s: open failed: %s\n"), fn,
		Fstrerror(fd));
	    return 1;
	}
	*fdp = fd;
	return 0;
    }

    /* no operation */
    if (*fdp != NULL && fn != NULL)
	return 0;

    /* XXX never reached */
    return 1;
}

/**
 * Copy header+payload, calculating digest(s) on the fly.
 * @param sfdp source file
 * @param sfnp source path
 * @param tfdp destination file
 * @param tfnp destination path
 */
static int copyFile(FD_t *sfdp, const char *sfnp,
		FD_t *tfdp, const char *tfnp)
{
    unsigned char buf[BUFSIZ];
    ssize_t count;
    int rc = 1;

    while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), *sfdp)) > 0)
    {
	if (Fwrite(buf, sizeof(buf[0]), count, *tfdp) != count) {
	    rpmlog(RPMLOG_ERR, _("%s: Fwrite failed: %s\n"), tfnp,
		Fstrerror(*tfdp));
	    goto exit;
	}
    }
    if (count < 0) {
	rpmlog(RPMLOG_ERR, _("%s: Fread failed: %s\n"), sfnp, Fstrerror(*sfdp));
	goto exit;
    }
    if (Fflush(*tfdp) != 0) {
	rpmlog(RPMLOG_ERR, _("%s: Fflush failed: %s\n"), tfnp,
	    Fstrerror(*tfdp));
    }

    rc = 0;

exit:
    return rc;
}

//------by weiliuhu 20240408
static int getRpmCfgKey()
{
	FILE *fp;
	int ret;
	char buf[128]={0};
	fp = fopen(RPMCFG, "rb");
	if(fp == NULL)
	{
		rpmlog_message(RPMLOG_WARNING, "rpmconfig.cfg file not exist!\n");
		return -1;
	}
	while(fgets(buf, 128, fp)){
		int len = strlen(buf);
		if(len > 0 && (buf[len-1]=='\n' || buf[len-1]=='\r'))
		{
			buf[--len]='\0';
		}
		if( !strcmp(buf, "rpmgmkey=1") )
		{
			ret = 0;	
			break;
		}				
		else{
			ret = -2;  //当存在配置文件，但value值不为1时，主要是为了让配置文件和value值都存在时，走国密，其他都按原来的走
		}
		memset(buf,0, sizeof(buf));
	}
	fclose(fp);
	return ret;
}

#include "rpmio/digest.h"
/*
 * Validate generated signature and insert to header if it looks sane.
 * NSS doesn't support everything GPG does. Basic tests to see if the 
 * generated signature is something we can use.
 * Return generated signature tag data on success, NULL on failure.
 */
static rpmtd makeSigTag(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen, int flag)
{
    pgpDigParams sigp = NULL;
    rpmTagVal sigtag;
    rpmtd sigtd = NULL;
    unsigned int hash_algo;
    unsigned int pubkey_algo;
    int ret = getRpmCfgKey();
    if(ret == 0)
	{
		if(flag == 1){
			sigtag = RPMSIGTAG_SM2;
		}
		else if(flag == 2){
			sigtag = RPMSIGTAG_SM3;
		}
		else{
			rpmlog_message(RPMLOG_DEBUG, "flag=%d\n" , flag);
			goto exit;
		}
	}
	else{
        if (pgpPrtParams(pkt, pktlen, PGPTAG_SIGNATURE, &sigp)) {
	    rpmlog(RPMLOG_ERR, _("Unsupported PGP signature\n"));
	    goto exit;
        }

        hash_algo = pgpDigParamsAlgo(sigp, PGPVAL_HASHALGO);
        if (rpmDigestLength(hash_algo) == 0) {
	    rpmlog(RPMLOG_ERR, _("Unsupported PGP hash algorithm %u\n"), hash_algo);
	    goto exit;
        }

        pubkey_algo = pgpDigParamsAlgo(sigp, PGPVAL_PUBKEYALGO);
        switch (pubkey_algo) {
        case PGPPUBKEYALGO_DSA:
	    sigtag = ishdr ? RPMSIGTAG_DSA : RPMSIGTAG_GPG;
	    break;
        case PGPPUBKEYALGO_RSA:
	    sigtag = ishdr ? RPMSIGTAG_RSA : RPMSIGTAG_PGP;
	    break;
        default:
	    rpmlog(RPMLOG_ERR, _("Unsupported PGP pubkey algorithm %u\n"),
		    pubkey_algo);
	    goto exit;
	    break;
        }
    }

    /* Looks sane, create the tag data */
    sigtd = rpmtdNew();
    sigtd->count = pktlen;
    sigtd->data = memcpy(xmalloc(pktlen), pkt, pktlen);;
    sigtd->type = RPM_BIN_TYPE;
    sigtd->tag = sigtag;
    sigtd->flags |= RPMTD_ALLOCED;

exit:
    if(sigp)
	{
        pgpDigParamsFree(sigp);
    }
    return sigtd;
}

static int runGPG(sigTarget sigt, const char *sigfile)
{
    int pid = 0, status;
    int pipefd[2];
    FILE *fpipe = NULL;
    unsigned char buf[BUFSIZ];
    ssize_t count;
    ssize_t wantCount;
    rpm_loff_t size;
    int rc = 1; /* assume failure */

    if (pipe(pipefd) < 0) {
        rpmlog(RPMLOG_ERR, _("Could not create pipe for signing: %m\n"));
        goto exit;
    }

    rpmPushMacro(NULL, "__plaintext_filename", NULL, "-", -1);
    rpmPushMacro(NULL, "__signature_filename", NULL, sigfile, -1);

    if (!(pid = fork())) {
	char *const *av;
	char *cmd = NULL;
	const char *tty = ttyname(STDIN_FILENO);
	const char *gpg_path = NULL;

	if (!getenv("GPG_TTY") && (!tty || setenv("GPG_TTY", tty, 0)))
	    rpmlog(RPMLOG_WARNING, _("Could not set GPG_TTY to stdin: %m\n"));

	gpg_path = rpmExpand("%{?_gpg_path}", NULL);
	if (gpg_path && *gpg_path != '\0')
	    (void) setenv("GNUPGHOME", gpg_path, 1);

	dup2(pipefd[0], STDIN_FILENO);
	close(pipefd[1]);

	unsetenv("MALLOC_CHECK_");
	cmd = rpmExpand("%{?__gpg_sign_cmd}", NULL);
	rc = poptParseArgvString(cmd, NULL, (const char ***)&av);
	if (!rc)
	    rc = execve(av[0], av+1, environ);

	rpmlog(RPMLOG_ERR, _("Could not exec %s: %s\n"), "gpg",
			strerror(errno));
	_exit(EXIT_FAILURE);
    }

    rpmPopMacro(NULL, "__plaintext_filename");
    rpmPopMacro(NULL, "__signature_filename");

    close(pipefd[0]);
    fpipe = fdopen(pipefd[1], "w");
    if (!fpipe) {
	rpmlog(RPMLOG_ERR, _("Could not open pipe for writing: %m\n"));
	goto exit;
    }

    if (Fseek(sigt->fd, sigt->start, SEEK_SET) < 0) {
	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		sigt->fileName, Fstrerror(sigt->fd));
	goto exit;
    }

    size = sigt->size;
    wantCount = size < sizeof(buf) ? size : sizeof(buf);
    while ((count = Fread(buf, sizeof(buf[0]), wantCount, sigt->fd)) > 0) {
	fwrite(buf, sizeof(buf[0]), count, fpipe);
	if (ferror(fpipe)) {
	    rpmlog(RPMLOG_ERR, _("Could not write to pipe\n"));
	    goto exit;
	}
	size -= count;
	wantCount = size < sizeof(buf) ? size : sizeof(buf);
    }
    if (count < 0) {
	rpmlog(RPMLOG_ERR, _("Could not read from file %s: %s\n"),
		sigt->fileName, Fstrerror(sigt->fd));
	goto exit;
    }

exit:

    if (fpipe)
	fclose(fpipe);
    if (pipefd[1])
	close(pipefd[1]);

    (void) waitpid(pid, &status, 0);
    pid = 0;
    if (!WIFEXITED(status) || WEXITSTATUS(status)) {
	rpmlog(RPMLOG_ERR, _("gpg exec failed (%d)\n"), WEXITSTATUS(status));
    } else {
	rc = 0;
    }
    return rc;
}

#include <gmssl/sm2.h>
#include <gmssl/x509_alg.h>
#include <gmssl/x509.h>
#include <gmssl/rand.h>
#include <gmssl/error.h>
//获取rpm文件被签名数据
static unsigned char *readFile(char *path, unsigned int *length) {
	FILE *fp;
	unsigned char *data;
	fp = fopen(path, "rb");
	if (fp == NULL) {
		rpmlog_message(RPMLOG_ERR, " path=%s\n",  path);
		return NULL;
	}
	//只获取rpm的文件部分数据
	fseek(fp, 96+8, SEEK_SET);
	uint8_t list[8];
	fread(list, 8, 1, fp);	
	int i;
	int x1;
	int x2;
	uint8_t k1[4] ={list[3],list[2],list[1], list[0]};
	
	memcpy(&x1, k1, 4);
	uint8_t k2[4] = {list[7],list[6],list[5],list[4]};
	
	memcpy(&x2, k2, 4);
	memset(list, 0, 8);
	
	//96+16+9*16+(4244+4)
	int newlen = 96+16+x1*16+x2+4 +8;
	
	fseek(fp, newlen, SEEK_SET);
	fread(list, 8, 1, fp);
	
	int entryNum;
	uint8_t m[4] ={list[3],list[2],list[1], list[0]};
	memcpy(&entryNum, m, 4);
	unsigned int dataNum;
	uint8_t t[4] = {list[7],list[6],list[5],list[4]};
	
	memcpy(&dataNum, t, 4);
	int offsetLen = 96+16+x1*16+(x2+4)+16 + entryNum*16 + dataNum;
	
	fseek(fp, 0, SEEK_END);
	unsigned int alllen = ftell(fp);
	unsigned int readlen = alllen - offsetLen;
	
	data = (unsigned char *)malloc((readlen + 1) * sizeof(char));
	fseek(fp, offsetLen, SEEK_SET);
	
	*length = fread(data, 1, readlen, fp);
	data[*length] = '\0';
	fclose(fp);
	return data;
}

//对rpm包签名
int kmsSignForRPM(char *priKeyPath, char *rpmbagPath, char *sigFile)
{
	int ret;
	SM2_KEY sm2_key;
	FILE *prikeyfp = NULL;
	prikeyfp = fopen(priKeyPath,"r");
	if(prikeyfp == NULL)
	{
		rpmlog_message(RPMLOG_WARNING, "fopen %s error!\n",  priKeyPath);
		return -1;
	}
	ret = sm2_private_key_info_decrypt_from_pem(&sm2_key,"123456",prikeyfp);
	if(ret != 1)
	{
		rpmlog_message(RPMLOG_WARNING, "sm2_private_key_info_decrypt_from_pem  error! ret=%d\n",  ret);
		fclose(prikeyfp);
		return -1;
	}
	fclose(prikeyfp);
	
	unsigned char dgst[32];
	FILE *certfp = NULL;
	unsigned char *bufcert;
	unsigned int filelen = 0;
	bufcert = readFile(rpmbagPath, &filelen);
	sm3_digest(bufcert, filelen, dgst);
	free(bufcert);
	
	unsigned char sig[1024]={0};
	size_t siglen;
	ret = sm2_sign(&sm2_key, dgst, sig, &siglen);
	if(ret != 1)
	{
		rpmlog_message(RPMLOG_ERR, " sm2_sign err!!! SM2 Signature: Failed!!!\n");
		return -1;
	}
	
	FILE *ffp = NULL;
	ffp = fopen(sigFile,"w");
	if(ffp == NULL)
	{
		rpmlog_message(RPMLOG_ERR, " fopen %s err!!!\n",  sigFile);
		return -1;
	}
	fwrite(sig, siglen, 1, ffp);
	fclose(ffp);
	rpmlog_message(RPMLOG_WARNING, "    SM2 Signature: Success!!! \n");
	return 0;
}	
//向sig文件内写入sm3处理后的hash值
static int kmsSM3ForRPM(char *sigFile,rpmtd sigtd_old)
{
	if(sigtd_old == NULL)
	{
		rpmlog_message(RPMLOG_ERR, " sigtd_old is null !!!\n");
		return -1;
	}
	unsigned char dgst[32];
	sm3_digest(sigtd_old->data, sigtd_old->count, dgst);
	FILE *ffp = NULL;
	ffp = fopen(sigFile,"w");
	if(ffp == NULL)
	{
		rpmlog_message(RPMLOG_ERR, " fopen %s err!!!\n",  sigFile);
		return -1;
	}
	fwrite(dgst, 32, 1, ffp);
	fclose(ffp);
	rpmlog_message(RPMLOG_WARNING, "    SM3 : Success!!! \n");
	return 0;
}

//-------------2024.03.05 by weiliuhu
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <gmssl/base64.h>

int doCmd(char *cmd)
{
    int ret = system(cmd);  
    if (ret == -1) 
     {
        rpmlog_message(RPMLOG_WARNING, "%s!!! system error!\n",  cmd);
        return -1;
     }
    if (WIFEXITED(ret))
     {
        if (0 == WEXITSTATUS(ret))
          {
            rpmlog_message(RPMLOG_WARNING, " run shell script %s successfully.\n",  cmd);
          }
        else{
            rpmlog_message(RPMLOG_WARNING, " run shell script %s fail, script exit code: %d\n",  cmd, WEXITSTATUS(ret));
            return -1;
          }
     }
    else{
        rpmlog_message(RPMLOG_WARNING, " %s exit ret = [%d]\n",  cmd, WEXITSTATUS(ret));
        return -1;
     }
    return 0;
}

static int mybase64_encode(uint8_t *buf, int in_len, uint8_t **dest, int *out_len)
{
    if(buf == NULL)
     {
        return -1;
     }
    BASE64_CTX ctx;
    int len;
    uint8_t *newstr = NULL;
    newstr = (uint8_t *)malloc(in_len*2);
    memset(newstr, 0, in_len*2);
    uint8_t *p = newstr;
    base64_encode_init(&ctx);
    base64_encode_update(&ctx, buf, in_len, p, &len); 

    p += len;	
    base64_encode_finish(&ctx, p , &len);

    p += len;	
    int newlen = p-newstr;
    
    *dest = newstr;
    //(*dest)[newlen]='\0';
    *out_len = newlen;
    return 0;
}

//对数据进行求sm3值，通过rpmguguard获取
static int kmsSM3ForRPMEx(char *sigFile,rpmtd sigtd_old)
{
	if(sigtd_old == NULL)
	{
		rpmlog_message(RPMLOG_ERR, " sigtd_old is null !!!\n");
		return -1;
	}
	int ret = 0;
	uint8_t *p = NULL;
	int outlen=0;
	ret = mybase64_encode(sigtd_old->data, sigtd_old->count, &p, &outlen);
	if(ret < 0)
	{
		rpmlog_message(RPMLOG_ERR, "mybase64_encode error!\n");
		return -1;
	}
    
	char *cmdbuf = NULL;
	cmdbuf = malloc(sizeof(char)*(1024+outlen));
	memset(cmdbuf, 0, 1024+outlen);
	sprintf(cmdbuf, "%s 2 \"%s\" \"%s\" %d %d", NFSSEC_GMPROXY, sigFile, p, outlen, sigtd_old->count);
	ret = doCmd(cmdbuf);
	if(ret < 0)
	{
		rpmlog_message(RPMLOG_ERR, "doCmd error!\n");
		if(p){
			free(p);
		}
		if(cmdbuf){
			free(cmdbuf);
		}
		return -1;
	}
	if(p){
		free(p);
	}
	if(cmdbuf){
		free(cmdbuf);
	}
	return 0;
}

/**
 * Generate GPG signature(s) for a header+payload file.
 * @param sigh		signature header
 * @param ishdr		header-only signature?
 * @param sigt		signature target
 * @param passPhrase	private key pass phrase
 * @return		generated sigtag on success, 0 on failure
 */
static rpmtd makeGPGSignature(Header sigh, int ishdr, sigTarget sigt, int flag, rpmtd sigtd_old)
{
    char * sigfile = rstrscat(NULL, sigt->fileName, ".sig", NULL);
    struct stat st;
    uint8_t * pkt = NULL;
    size_t pktlen = 0;
    rpmtd sigtd = NULL;
    int ret = getRpmCfgKey();
    if(ret == 0)
	{
		//对rpm数据求sm2签名数据
		if(flag == 1){
			//走签名进程进行写sig文件   2024.03.05 by weiliuhu
			char cmdbuf[1024]={0};
			sprintf(cmdbuf, "%s 1 \"%s\" \"%s\" \"%s\"", NFSSEC_GMPROXY, RPMPRIKEY, sigt->fileName, sigfile);
			ret = doCmd(cmdbuf);
			if(ret < 0)
			{
				rpmlog_message(RPMLOG_ERR, "error doCmd\n" );
				goto exit;
			}
		}
		else if(flag == 2){
			if(kmsSM3ForRPMEx(sigfile, sigtd_old) < 0)   //走签名进程进行写sig文件   2024.03.05 by weiliuhu
			{
				rpmlog_message(RPMLOG_ERR, "error kmsSM3ForRPMEx\n" );
				goto exit;
			}
		}
		else{
			rpmlog_message(RPMLOG_ERR, "  flag=%d\n" ,flag);
			goto exit;
		}
	}
	else{
		if (runGPG(sigt, sigfile))
		{
			rpmlog(RPMLOG_ERR, "runGPG failed\n" );
			goto exit;
		}
	}

    if (stat(sigfile, &st)) {
	/* GPG failed to write signature */
	rpmlog(RPMLOG_ERR, _("gpg failed to write signature\n"));
	goto exit;
    }

    pktlen = st.st_size;
    rpmlog(RPMLOG_DEBUG, "GPG sig size: %zd\n", pktlen);
    pkt = xmalloc(pktlen);

    {	FD_t fd;

	int rc = 0;
	fd = Fopen(sigfile, "r.ufdio");
	if (fd != NULL && !Ferror(fd)) {
	    rc = Fread(pkt, sizeof(*pkt), pktlen, fd);
	    (void) Fclose(fd);
	}
	if (rc != pktlen) {
	    rpmlog(RPMLOG_ERR, _("unable to read the signature\n"));
	    goto exit;
	}
    }

    rpmlog(RPMLOG_DEBUG, "Got %zd bytes of GPG sig\n", pktlen);

    /* Parse the signature, change signature tag as appropriate. */
    sigtd = makeSigTag(sigh, ishdr, pkt, pktlen, flag);
exit:
    (void) unlink(sigfile);
    free(sigfile);
    free(pkt);

    return sigtd;
}

static void deleteSigs(Header sigh)
{
    headerDel(sigh, RPMSIGTAG_GPG);
    headerDel(sigh, RPMSIGTAG_PGP);
    headerDel(sigh, RPMSIGTAG_DSA);
    headerDel(sigh, RPMSIGTAG_RSA);
    headerDel(sigh, RPMSIGTAG_PGP5);
    headerDel(sigh, RPMSIGTAG_SM2);
    headerDel(sigh, RPMSIGTAG_SM3);
}

static int haveSignature(rpmtd sigtd, Header h)
{
    pgpDigParams sig1 = NULL;
    pgpDigParams sig2 = NULL;
    struct rpmtd_s oldtd;
    int rc = 0; /* assume no */

    if (!headerGet(h, rpmtdTag(sigtd), &oldtd, HEADERGET_DEFAULT))
	return rc;

    pgpPrtParams(sigtd->data, sigtd->count, PGPTAG_SIGNATURE, &sig1);
    while (rpmtdNext(&oldtd) >= 0 && rc == 0) {
	pgpPrtParams(oldtd.data, oldtd.count, PGPTAG_SIGNATURE, &sig2);
	if (pgpDigParamsCmp(sig1, sig2) == 0)
	    rc = 1;
	sig2 = pgpDigParamsFree(sig2);
    }
    pgpDigParamsFree(sig1);
    rpmtdFreeData(&oldtd);

    return rc;
}
//判断rpm文件内有无国密算法标记
static int haveSignatureForSM(rpmtd sigtd, Header h)
{
	int ret;
	for(int i = 0; i< ((GMHeader)h)->indexUsed; i++)
	{
		if( ((GMHeader)h)->index[i].info.tag == sigtd->tag ) // 如有sm2加密的，就算是已经被加密&& (h->index[i].info.type == sigtd->type) && (h->index[i].info.count == sigtd->count) )
		{
			return 1;
		}
	}
	return 0;
}

//add sm2 sign by weiliuhu 2024.01.26
static int replaceSignature(Header sigh, sigTarget sigt_v3, sigTarget sigt_v4)
{
    int rc = -1;
    rpmtd sigtd = NULL;
    rpmtd sigtd2 = NULL;
    
    /* Make the cheaper v4 signature first */
    if ((sigtd = makeGPGSignature(sigh, 1, sigt_v4,1,NULL)) == NULL)
	goto exit;

    /* See if we already have a signature by the same key and parameters */
    int ret = getRpmCfgKey();
    if(ret == 0){
    	if(haveSignatureForSM(sigtd, sigh)){
    		rpmlog(RPMLOG_WARNING, "gm have encrypt!!!\n");
			rc = 1;
			goto exit;
    	}
     }
    else{
    	if (haveSignature(sigtd, sigh)) {
	    	rpmlog(RPMLOG_WARNING, "gpg have encrypt!!!\n");
			rc = 1;
			goto exit;
		}
     }
    /* Nuke all signature tags */
    deleteSigs(sigh);

    if (headerPut(sigh, sigtd, HEADERPUT_DEFAULT) == 0)
	goto exit;
    //rpmtdFree(sigtd);         //暂时屏蔽，修改为2个变量，在使用国密时，可以用上一次生成的签名数据直接拿来sm3

    /* Assume the same signature test holds for v3 signature too */
    if ((sigtd2 = makeGPGSignature(sigh, 0, sigt_v3, 2, sigtd)) == NULL)
	goto exit;

    if (headerPut(sigh, sigtd2, HEADERPUT_DEFAULT) == 0)
	goto exit;

    rc = 0;
exit:
    rpmtdFree(sigtd);
    rpmtdFree(sigtd2);
    return rc;
}

static void unloadImmutableRegion(Header *hdrp, rpmTagVal tag)
{
    struct rpmtd_s td;
    rpmtd utd = &td;
    Header nh;
    Header oh;

    if (headerGet(*hdrp, tag, utd, HEADERGET_DEFAULT)) {
	oh = headerCopyLoad(utd->data);
	rpmtdFreeData(utd);
    } else {
	/* XXX should we warn if the immutable region is corrupt/missing? */
	oh = headerLink(*hdrp);
    }

    if (oh) {
	/* Perform a copy to eliminate crud from buggy signing tools etc */
	nh = headerCopy(oh);
	headerFree(*hdrp);
	*hdrp = headerLink(nh);
	headerFree(nh);
	headerFree(oh);
    }
}

#ifdef WITH_IMAEVM
static rpmRC replaceSigDigests(FD_t fd, const char *rpm, Header *sigp,
			       off_t sigStart, off_t sigTargetSize,
			       char *SHA256, char *SHA1, uint8_t *MD5)
{
    off_t archiveSize;
    rpmRC rc = RPMRC_OK;

    if (Fseek(fd, sigStart, SEEK_SET) < 0) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		rpm, Fstrerror(fd));
	goto exit;
    }

    /* Get payload size from signature tag */
    archiveSize = headerGetNumber(*sigp, RPMSIGTAG_PAYLOADSIZE);
    if (!archiveSize) {
	archiveSize = headerGetNumber(*sigp, RPMSIGTAG_LONGARCHIVESIZE);
    }

    /* Set reserved space to 0 */
    rpmPushMacro(NULL, "__gpg_reserved_space", NULL, 0, RMIL_GLOBAL);

    /* Replace old digests in sigh */
    rc = rpmGenerateSignature(SHA256, SHA1, MD5, sigTargetSize, archiveSize, fd);
    if (rc != RPMRC_OK) {
	rpmlog(RPMLOG_ERR, _("generateSignature failed\n"));
	goto exit;
    }

    if (Fseek(fd, sigStart, SEEK_SET) < 0) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		rpm, Fstrerror(fd));
	goto exit;
    }

    headerFree(*sigp);
    rc = rpmReadSignature(fd, sigp, NULL);
    if (rc != RPMRC_OK) {
	rpmlog(RPMLOG_ERR, _("rpmReadSignature failed\n"));
	goto exit;
    }

exit:
    return rc;
}
#endif

static rpmRC includeFileSignatures(FD_t fd, const char *rpm,
				   Header *sigp, Header *hdrp,
				   off_t sigStart, off_t headerStart)
{
#ifdef WITH_IMAEVM
    FD_t ofd = NULL;
    char *trpm = NULL;
    char *key;
    char *keypass;
    char *SHA1 = NULL;
    char *SHA256 = NULL;
    uint8_t *MD5 = NULL;
    off_t sigTargetSize;
    rpmRC rc = RPMRC_OK;
    struct rpmtd_s osigtd;
    char *o_sha1 = NULL;

    unloadImmutableRegion(hdrp, RPMTAG_HEADERIMMUTABLE);

    key = rpmExpand("%{?_file_signing_key}", NULL);

    keypass = rpmExpand("%{?_file_signing_key_password}", NULL);
    if (rstreq(keypass, "")) {
	free(keypass);
	keypass = NULL;
    }

    rc = rpmSignFiles(*hdrp, key, keypass);
    if (rc != RPMRC_OK) {
	goto exit;
    }

    *hdrp = headerReload(*hdrp, RPMTAG_HEADERIMMUTABLE);
    if (*hdrp == NULL) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("headerReload failed\n"));
	goto exit;
    }

    ofd = rpmMkTempFile(NULL, &trpm);
    if (ofd == NULL || Ferror(ofd)) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("rpmMkTemp failed\n"));
	goto exit;
    }

    /* Copy archive to temp file */
    if (copyFile(&fd, rpm, &ofd, trpm)) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("copyFile failed\n"));
	goto exit;
    }

    if (Fseek(fd, headerStart, SEEK_SET) < 0) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		rpm, Fstrerror(fd));
	goto exit;
    }

    /* Start MD5 calculation */
    fdInitDigestID(fd, PGPHASHALGO_MD5, RPMSIGTAG_MD5, 0);

    /* Write header to rpm and recalculate digests */
    fdInitDigestID(fd, PGPHASHALGO_SHA1, RPMSIGTAG_SHA1, 0);
    fdInitDigestID(fd, PGPHASHALGO_SHA256, RPMSIGTAG_SHA256, 0);
    rc = headerWrite(fd, *hdrp, HEADER_MAGIC_YES);
    if (rc != RPMRC_OK) {
	rpmlog(RPMLOG_ERR, _("headerWrite failed\n"));
	goto exit;
    }
    fdFiniDigest(fd, RPMSIGTAG_SHA1, (void **)&SHA1, NULL, 1);
    /* Only add SHA256 if it was there to begin with */
    if (headerIsEntry(*sigp, RPMSIGTAG_SHA256))
	fdFiniDigest(fd, RPMSIGTAG_SHA256, (void **)&SHA256, NULL, 1);

    /* Copy archive from temp file */
    if (Fseek(ofd, 0, SEEK_SET) < 0) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		rpm, Fstrerror(fd));
	goto exit;
    }
    if (copyFile(&ofd, trpm, &fd, rpm)) {
	rc = RPMRC_FAIL;
	rpmlog(RPMLOG_ERR, _("copyFile failed\n"));
	goto exit;
    }
    unlink(trpm);

    sigTargetSize = Ftell(fd) - headerStart;
    fdFiniDigest(fd, RPMSIGTAG_MD5, (void **)&MD5, NULL, 0);

    if (headerGet(*sigp, RPMSIGTAG_SHA1, &osigtd, HEADERGET_DEFAULT)) {
	o_sha1 = xstrdup(osigtd.data);
	rpmtdFreeData(&osigtd);
    }

    if (strcmp(SHA1, o_sha1) == 0)
	rpmlog(RPMLOG_WARNING,
	       _("%s already contains identical file signatures\n"),
	       rpm);
    else
	replaceSigDigests(fd, rpm, sigp, sigStart, sigTargetSize, SHA256, SHA1, MD5);

exit:
    free(trpm);
    free(MD5);
    free(SHA1);
    free(SHA256);
    free(o_sha1);
    free(keypass);
    free(key);
    if (ofd)
	(void) closeFile(&ofd);
    return rc;
#else
    rpmlog(RPMLOG_ERR, _("file signing support not built in\n"));
    return RPMRC_FAIL;
#endif
}

static int msgCb(struct rpmsinfo_s *sinfo, void *cbdata)
{
    char **msg = cbdata;
    if (sinfo->rc && *msg == NULL)
	*msg = rpmsinfoMsg(sinfo);
    return (sinfo->rc != RPMRC_FAIL);
}

/* Require valid digests on entire package for signing. */
static int checkPkg(FD_t fd, char **msg)
{
    int rc;
    struct rpmvs_s *vs = rpmvsCreate(RPMSIG_DIGEST_TYPE, 0, NULL);
    off_t offset = Ftell(fd);

    Fseek(fd, 0, SEEK_SET);
    rc = rpmpkgRead(vs, fd, NULL, NULL, msg);
    if (!rc)
	rc = rpmvsVerify(vs, RPMSIG_DIGEST_TYPE, msgCb, msg);
    Fseek(fd, offset, SEEK_SET);

    rpmvsFree(vs);
    return rc;
}

/** \ingroup rpmcli
 * Create/modify elements in signature header.
 * @param rpm		path to package
 * @param deleting	adding or deleting signature?
 * @param signfiles	sign files if non-zero
 * @return		0 on success, -1 on error
 */
static int rpmSign(const char *rpm, int deleting, int signfiles)
{
    FD_t fd = NULL;
    FD_t ofd = NULL;
    char *trpm = NULL;
    Header sigh = NULL;
    Header h = NULL;
    char *msg = NULL;
    int res = -1; /* assume failure */
    rpmRC rc;
    struct rpmtd_s utd;
    off_t headerStart;
    off_t sigStart;
    struct sigTarget_s sigt_v3;
    struct sigTarget_s sigt_v4;
    unsigned int origSigSize;
    int insSig = 0;

    fprintf(stdout, "%s:\n", rpm);

    if (manageFile(&fd, rpm, O_RDWR))
	goto exit;

    /* Ensure package is intact before attempting to sign */
    if ((rc = checkPkg(fd, &msg))) {
	rpmlog(RPMLOG_ERR, "not signing corrupt package %s: %s\n", rpm, msg);
	goto exit;
    }

    if ((rc = rpmLeadRead(fd, &msg)) != RPMRC_OK) {
	rpmlog(RPMLOG_ERR, "%s: %s\n", rpm, msg);
	goto exit;
    }

    sigStart = Ftell(fd);
    rc = rpmReadSignature(fd, &sigh, &msg);
    if (rc != RPMRC_OK) {
	rpmlog(RPMLOG_ERR, _("%s: rpmReadSignature failed: %s"), rpm,
		    (msg && *msg ? msg : "\n"));
	goto exit;
    }

    headerStart = Ftell(fd);
    if (rpmReadHeader(NULL, fd, &h, &msg) != RPMRC_OK) {
	rpmlog(RPMLOG_ERR, _("%s: headerRead failed: %s\n"), rpm, msg);
	goto exit;
    }

    if (!headerIsEntry(h, RPMTAG_HEADERIMMUTABLE)) {
	rpmlog(RPMLOG_ERR, _("Cannot sign RPM v3 packages\n"));
	goto exit;
    }

    if (signfiles) {
	includeFileSignatures(fd, rpm, &sigh, &h, sigStart, headerStart);
    }

    unloadImmutableRegion(&sigh, RPMTAG_HEADERSIGNATURES);
    origSigSize = headerSizeof(sigh, HEADER_MAGIC_YES);

    if (deleting) {	/* Nuke all the signature tags. */
	deleteSigs(sigh);
    } else {
	/* Signature target containing header + payload */
	sigt_v3.fd = fd;
	sigt_v3.start = headerStart;
	sigt_v3.fileName = rpm;
	sigt_v3.size = fdSize(fd) - headerStart;

	/* Signature target containing only header */
	sigt_v4 = sigt_v3;
	sigt_v4.size = headerSizeof(h, HEADER_MAGIC_YES);

	res = replaceSignature(sigh, &sigt_v3, &sigt_v4);
	if (res != 0) {
	    if (res == 1) {
		rpmlog(RPMLOG_WARNING,
		   _("%s already contains identical signature, skipping\n"),
		   rpm);
		/* Identical signature is not an error */
		res = 0;
	    }
		else{
			rpmlog(RPMLOG_ERR, 
		   _("%s sign err\n"),
		   rpm);
		}
	    goto exit;
	}
	res = -1;
    }

    /* Try to make new signature smaller to have size of original signature */
    rpmtdReset(&utd);
    if (headerGet(sigh, RPMSIGTAG_RESERVEDSPACE, &utd, HEADERGET_MINMEM)) {
	int diff;
	int count;
	char *reservedSpace = NULL;

	count = utd.count;
	diff = headerSizeof(sigh, HEADER_MAGIC_YES) - origSigSize;

	if (diff < count) {
	    reservedSpace = xcalloc(count - diff, sizeof(char));
	    headerDel(sigh, RPMSIGTAG_RESERVEDSPACE);
	    rpmtdReset(&utd);
	    utd.tag = RPMSIGTAG_RESERVEDSPACE;
	    utd.count = count - diff;
	    utd.type = RPM_BIN_TYPE;
	    utd.data = reservedSpace;
	    headerPut(sigh, &utd, HEADERPUT_DEFAULT);
	    free(reservedSpace);
	    insSig = 1;
	}
    }

    /* Reallocate the signature into one contiguous region. */
    sigh = headerReload(sigh, RPMTAG_HEADERSIGNATURES);
    if (sigh == NULL)	/* XXX can't happen */
	goto exit;

    if (insSig) {
	/* Insert new signature into original rpm */
	if (Fseek(fd, sigStart, SEEK_SET) < 0) {
	    rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		    rpm, Fstrerror(fd));
	    goto exit;
	}

	if (rpmWriteSignature(fd, sigh)) {
	    rpmlog(RPMLOG_ERR, _("%s: rpmWriteSignature failed: %s\n"), rpm,
		Fstrerror(fd));
	    goto exit;
	}
	res = 0;
    } else {
	/* Replace orignal rpm with new rpm containing new signature */
	rasprintf(&trpm, "%s.XXXXXX", rpm);
	ofd = rpmMkTemp(trpm);
	if (ofd == NULL || Ferror(ofd)) {
	    rpmlog(RPMLOG_ERR, _("rpmMkTemp failed\n"));
	    goto exit;
	}

	/* Write the lead/signature of the output rpm */
	rc = rpmLeadWrite(ofd, h);
	if (rc != RPMRC_OK) {
	    rpmlog(RPMLOG_ERR, _("%s: writeLead failed: %s\n"), trpm,
		Fstrerror(ofd));
	    goto exit;
	}

	if (rpmWriteSignature(ofd, sigh)) {
	    rpmlog(RPMLOG_ERR, _("%s: rpmWriteSignature failed: %s\n"), trpm,
		Fstrerror(ofd));
	    goto exit;
	}

	if (Fseek(fd, headerStart, SEEK_SET) < 0) {
	    rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
		    rpm, Fstrerror(fd));
	    goto exit;
	}
	/* Append the header and archive from the temp file */
	if (copyFile(&fd, rpm, &ofd, trpm) == 0) {
	    struct stat st;

	    /* Move final target into place, restore file permissions. */
	    if (stat(rpm, &st) == 0 && unlink(rpm) == 0 &&
			rename(trpm, rpm) == 0 && chmod(rpm, st.st_mode) == 0) {
		res = 0;
	    } else {
		rpmlog(RPMLOG_ERR, _("replacing %s failed: %s\n"),
		       rpm, strerror(errno));
	    }
	}
    }

exit:
    if (fd)	(void) closeFile(&fd);
    if (ofd)	(void) closeFile(&ofd);

    headerFree(sigh);
    headerFree(h);
    free(msg);

    /* Clean up intermediate target */
    if (trpm) {
	(void) unlink(trpm);
	free(trpm);
    }

    return res;
}

int rpmPkgSign(const char *path, const struct rpmSignArgs * args)
{
    int rc;

    if (args) {
	if (args->hashalgo) {
	    char *algo = NULL;
	    rasprintf(&algo, "%d", args->hashalgo);
	    rpmPushMacro(NULL, "_gpg_digest_algo", NULL, algo, RMIL_GLOBAL);
	    free(algo);
	}
	if (args->keyid) {
	    rpmPushMacro(NULL, "_gpg_name", NULL, args->keyid, RMIL_GLOBAL);
	}
    }

    rc = rpmSign(path, 0, args ? args->signfiles : 0);

    if (args) {
	if (args->hashalgo) {
	    rpmPopMacro(NULL, "_gpg_digest_algo");
	}
	if (args->keyid) {
	    rpmPopMacro(NULL, "_gpg_name");
	}
    }

    return rc;
}

int rpmPkgDelSign(const char *path, const struct rpmSignArgs * args)
{
    return rpmSign(path, 1, 0);
}
